"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstance = exports.getServiceProvider = void 0;
const metadata_1 = require("../common/metadata");
const constants_1 = require("../common/constants");
const Errors_1 = require("../system/Errors");
const AbstractLogger_1 = require("../logging/AbstractLogger");
const isSubClass = (derivedClass, baseClass) => {
    return derivedClass === baseClass || derivedClass.prototype instanceof baseClass;
};
class DefaultServiceInjector {
    constructor(injectableRegistry, providerRegistry) {
        this.injectableRegistry = injectableRegistry;
        this.providerRegistry = providerRegistry;
        this.initables = [];
        this.assertNoCircularDependency();
    }
    initInstances() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const initable of this.initables) {
                yield initable.init();
            }
            (0, exports.getServiceProvider)().addInitables(this.initables);
            this.initables = [];
        });
    }
    resolveMultiple(abstractConstructorFn) {
        const providers = this.providerRegistry.filter((reg) => isSubClass(reg.baseClass, abstractConstructorFn));
        const injectables = this.injectableRegistry.filter((reg) => isSubClass(reg.constructorFn, abstractConstructorFn));
        if (!providers.length && !injectables.length) {
            return [];
        }
        //-- prioritize provider
        if (providers.length) {
            return providers.map((providerInfo) => {
                const provider = this.resolve(providerInfo.provider);
                let instance;
                if (providerInfo.mode !== "singleton" || !providerInfo.singletonInstance) {
                    instance = provider.provide();
                    if (providerInfo.mode === "singleton") {
                        providerInfo.singletonInstance = instance;
                    }
                    const metadata = (0, metadata_1.getObjectMetadata)(instance.constructor);
                    if (metadata === null || metadata === void 0 ? void 0 : metadata.initable) {
                        this.initables.push(instance);
                    }
                }
                else {
                    instance = providerInfo.singletonInstance;
                }
                return instance;
            });
        }
        if (injectables.length) {
            return injectables.map((reg) => {
                switch (reg.mode) {
                    case "singleton":
                        if (!reg.singletonInstance) {
                            reg.singletonInstance = this.getNewInstance(reg);
                        }
                        return reg.singletonInstance;
                    case "transient":
                        return this.getNewInstance(reg);
                }
            });
        }
        return [];
    }
    resolveOptional(abstractConstructorFn) {
        const instances = this.resolveMultiple(abstractConstructorFn);
        if (!instances.length) {
            return undefined;
        }
        if (instances.length > 1) {
            throw Errors_1.ErrorBuilder.forMutipleInjectionsFound(abstractConstructorFn, instances.map((instance) => instance.constructor));
        }
        return instances[0];
    }
    resolve(abstractConstructorFn) {
        const instances = this.resolveMultiple(abstractConstructorFn);
        if (!instances.length) {
            throw Errors_1.ErrorBuilder.forInjectionNotFound(abstractConstructorFn);
        }
        if (instances.length > 1) {
            throw Errors_1.ErrorBuilder.forMutipleInjectionsFound(abstractConstructorFn, instances.map((instance) => instance.constructor));
        }
        return instances[0];
    }
    getNewInstance(info) {
        const dependencies = (info.dependencies || []).map((dep) => dep.required ? this.resolve(dep.ctor) : this.resolveOptional(dep.ctor));
        const instance = new info.constructorFn(...dependencies);
        const metadata = (0, metadata_1.getObjectMetadata)(info.constructorFn);
        if (metadata === null || metadata === void 0 ? void 0 : metadata.initable) {
            this.initables.push(instance);
        }
        if (metadata === null || metadata === void 0 ? void 0 : metadata.logContext) {
            dependencies.forEach((dep) => {
                if (dep instanceof AbstractLogger_1.AbstractLogger) {
                    dep[constants_1.LOG_CONTEXT] = metadata.logContext;
                }
            });
        }
        return instance;
    }
    assertNoCircularDependency() {
        //-- build dag nodes
        let dagNodes = this.injectableRegistry.map((reg) => ({
            constructorFn: reg.constructorFn,
            dependencies: (reg.dependencies || []).slice(),
        }));
        let removeCount = dagNodes.length;
        while (removeCount > 0) {
            //-- remove nodes that has empty dependency until can no more remove
            const tobeRemoved = dagNodes.filter((node) => 
            //-- no dependency
            !node.dependencies.length ||
                //-- or every dependency has no candidate
                node.dependencies.every((dep) => !dagNodes.some((otherNode) => isSubClass(otherNode.constructorFn, dep.ctor))));
            removeCount = tobeRemoved.length;
            //-- remove tobeRemoved nodes from dagNodes
            dagNodes = dagNodes.filter((node) => !tobeRemoved.includes(node));
            //-- remove the tobeRemoved nodes from dependencies list of remaining nodes
            dagNodes.forEach((node) => {
                node.dependencies = node.dependencies.filter((dep) => !tobeRemoved.find((removeNode) => isSubClass(removeNode.constructorFn, dep.ctor)));
            });
        }
        if (dagNodes.length) {
            throw Errors_1.ErrorBuilder.forInjectionCircularDependency(dagNodes.map((dag) => dag.constructorFn));
        }
    }
}
class ServiceProvider {
    constructor() {
        this.injectableRegistry = [];
        this.providerRegistry = [];
        this.initables = [];
    }
    //@internal
    addInitables(initables) {
        this.initables = this.initables.concat(initables);
    }
    //@internal
    getInitables() {
        return this.initables;
    }
    getInfo(constructorFn) {
        return this.injectableRegistry.find((reg) => isSubClass(reg.constructorFn, constructorFn));
    }
    register(constructorFn, mode) {
        var _a;
        const index = this.injectableRegistry.findIndex((reg) => reg.constructorFn === constructorFn);
        const reg = {
            constructorFn,
            mode: mode || "singleton",
            dependencies: ((_a = (0, metadata_1.getObjectMetadata)(constructorFn)) === null || _a === void 0 ? void 0 : _a.constructorDependencies) ||
                (Reflect.getMetadata("design:paramtypes", constructorFn) || []).map((ctor) => ({
                    ctor,
                    required: true,
                })),
        };
        if (index < 0) {
            this.injectableRegistry.push(reg);
        }
        else {
            this.injectableRegistry[index] = Object.assign(Object.assign({}, this.injectableRegistry[index]), reg);
        }
        return this;
    }
    addProvider(baseClass, provider) {
        const index = this.providerRegistry.findIndex((reg) => reg.baseClass === baseClass);
        const metadata = (0, metadata_1.getObjectMetadata)(provider);
        const reg = { baseClass, provider, mode: (metadata === null || metadata === void 0 ? void 0 : metadata.injectionMode) || "singleton" };
        if (index < 0) {
            this.providerRegistry.push(reg);
        }
        else {
            this.providerRegistry[index] = reg;
        }
        return this;
    }
    getInjector() {
        return new DefaultServiceInjector(this.injectableRegistry.slice(), this.providerRegistry.slice());
    }
}
const getServiceProvider = () => {
    let serviceProvider = globalThis[constants_1.GLOBAL_SERVICE_PROVIDER];
    if (!serviceProvider) {
        serviceProvider = new ServiceProvider();
        globalThis[constants_1.GLOBAL_SERVICE_PROVIDER] = serviceProvider;
    }
    return serviceProvider;
};
exports.getServiceProvider = getServiceProvider;
const getInstance = (cls) => __awaiter(void 0, void 0, void 0, function* () {
    const injector = (0, exports.getServiceProvider)().getInjector();
    const instance = injector.resolve(cls);
    yield injector.initInstances();
    return instance;
});
exports.getInstance = getInstance;
