"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HasMany = exports.HasOne = exports.ElementDto = exports.Enum = exports.Relation = exports.Field = exports.Data = void 0;
const DataType_1 = require("../common/DataType");
const metadata_1 = require("../common/metadata");
const Errors_1 = require("../system/Errors");
/**
 * A simple data transfer object.
 */
const Data = (config) => (constructor) => {
    const metadata = (0, metadata_1.initObjectMetadata)(constructor.prototype);
    metadata.relaxed = config === null || config === void 0 ? void 0 : config.relaxed;
    metadata.description = config === null || config === void 0 ? void 0 : config.description;
    metadata.ignoreCrud = config === null || config === void 0 ? void 0 : config.ignoreCrud;
    //-- try create sample model and check for default values
    const sampleModel = new constructor();
    const defaultFields = Object.keys(sampleModel).filter((key) => metadata.fields.find((f) => f.name === key));
    defaultFields.forEach((defaultField) => {
        const field = (0, metadata_1.initFieldMetadata)(constructor.prototype, defaultField);
        field.defaultValue = sampleModel[defaultField];
    });
    return constructor;
};
exports.Data = Data;
const Field = (config) => (prototype, propertyKey) => {
    const field = (0, metadata_1.initFieldMetadata)(prototype, propertyKey);
    Object.assign(field, config);
    //-- properties interpolation
    if (field.isTimestamp) {
        field.isBigInt = true;
        field.rangeProps = { min: 0 };
    }
    if (field.vectorProps) {
        field.dataType = DataType_1.DataType.ARRAY;
    }
    else if (field.elementDto) {
        field.dataType = DataType_1.DataType.OBJECT;
    }
    //-- properties validation
    if (field.mimeProps && field.dataType !== DataType_1.DataType.STRING) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Cannot use uri on non-text field of ${prototype.constructor.name}:${propertyKey}`);
    }
    if (field.isBigInt && field.dataType !== DataType_1.DataType.NUMBER) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Cannot use big int on non-numeric field of ${prototype.constructor.name}:${propertyKey}`);
    }
    if (field.textLength && (field.textLength <= 0 || !Number.isInteger(field.textLength))) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Text length only accepts integer value, provided: ${field.textLength}`);
    }
};
exports.Field = Field;
const Relation = (model, cascadeType) => {
    const metadata = (0, metadata_1.getObjectMetadata)(model);
    if (!metadata) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Cannot find dto metadata: ${model.name}`);
    }
    return { fk: { dtoId: metadata.id, cascade: cascadeType } };
};
exports.Relation = Relation;
const Enum = (enumClass) => {
    const values = Object.values(enumClass);
    const numberValues = values.filter((v) => typeof v === "number");
    const stringValues = values.filter((v) => typeof v === "string");
    return { enum: numberValues.length ? numberValues : stringValues };
};
exports.Enum = Enum;
const ElementDto = (dtoClass) => {
    return { elementDto: (0, metadata_1.getObjectMetadata)(dtoClass) };
};
exports.ElementDto = ElementDto;
const Has = (config) => (prototype, propertyKey) => {
    const dtoMedata = (0, metadata_1.getObjectMetadata)(config.model);
    const field = (0, metadata_1.initFieldMetadata)(prototype, propertyKey);
    field.vectorProps = config.single ? undefined : Object.assign(Object.assign({}, field.vectorProps), { elementDataType: DataType_1.DataType.OBJECT });
    field.hasMany = {
        relationDto: dtoMedata,
        column: config.column,
        single: config.single,
    };
    //-- init FK for target model
    const targetField = (0, metadata_1.initFieldMetadata)(config.model.prototype, config.column);
    targetField.fk = (0, exports.Relation)(prototype.constructor, config.cascade || "delete").fk;
};
const HasOne = (config) => (prototype, propertyKey) => {
    return Has(Object.assign(Object.assign({}, config), { single: true }))(prototype, propertyKey);
};
exports.HasOne = HasOne;
const HasMany = (config) => (prototype, propertyKey) => {
    return Has(Object.assign(Object.assign({}, config), { single: false }))(prototype, propertyKey);
};
exports.HasMany = HasMany;
