"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FK = exports.Column = exports.Index = exports.Model = void 0;
const metadata_1 = require("../common/metadata");
const system_1 = require("../system/system");
const decorator_1 = require("../dto/decorator");
const Errors_1 = require("../system/Errors");
const registerModel = (model) => {
    const store = (0, system_1.getGlobalStore)();
    if (!store.models) {
        store.models = [];
    }
    const models = store.models;
    const modelMetadata = (0, metadata_1.getObjectMetadata)(model);
    const duplicatedModelIndex = models.findIndex((m) => {
        const metadata = (0, metadata_1.getObjectMetadata)(m);
        return (metadata === null || metadata === void 0 ? void 0 : metadata.id) === (modelMetadata === null || modelMetadata === void 0 ? void 0 : modelMetadata.id);
    });
    if (duplicatedModelIndex >= 0) {
        models[duplicatedModelIndex] = model;
    }
    else {
        //-- check if this model has @HasMany on it, if so find the appropriate position
        const hasManyFields = (modelMetadata === null || modelMetadata === void 0 ? void 0 : modelMetadata.fields.filter((f) => f.hasMany)) || [];
        if (!hasManyFields.length) {
            //-- no has many fields, just add to array because it complies with DAG order
            models.push(model);
        }
        else {
            //-- this model contains other models as @HasMany / @HasOne, find the position of the first child model and put it right before
            const firstChildPosition = models.findIndex((m) => {
                const childMeta = (0, metadata_1.getObjectMetadata)(m);
                return hasManyFields.find((f) => { var _a; return ((_a = f.hasMany) === null || _a === void 0 ? void 0 : _a.relationDto.id) === (childMeta === null || childMeta === void 0 ? void 0 : childMeta.id); });
            });
            if (firstChildPosition >= 0) {
                //-- insert to array at firstChildPosition
                models.splice(firstChildPosition, 0, model);
            }
            else {
                //-- just insert it normally
                models.push(model);
            }
        }
    }
};
const Model = (config) => (constructor) => {
    const meta = (0, metadata_1.initObjectMetadata)(constructor.prototype);
    meta.id = (0, metadata_1.getTableName)((config === null || config === void 0 ? void 0 : config.override) || constructor);
    //-- register model here
    registerModel(constructor);
    return (0, decorator_1.Data)({ description: config === null || config === void 0 ? void 0 : config.description, ignoreCrud: config === null || config === void 0 ? void 0 : config.ignoreCrud })(constructor);
};
exports.Model = Model;
const Index = (fields, config) => (constructor) => {
    const meta = (0, metadata_1.initObjectMetadata)(constructor.prototype);
    if (!meta.indexes) {
        meta.indexes = [];
    }
    meta.indexes.push({
        columns: fields,
        sparse: config === null || config === void 0 ? void 0 : config.sparse,
        unique: config === null || config === void 0 ? void 0 : config.unique,
    });
};
exports.Index = Index;
const Column = (props) => (prototype, propertyKey) => {
    const field = (0, metadata_1.initFieldMetadata)(prototype, propertyKey);
    Object.assign(field, props);
    if (field.pk) {
        field.isRequired = true;
        if (field.pk.isAutoIncrement) {
            field.serverValue = true;
        }
    }
    if (field.creationTimestamp || field.modificationTimestamp) {
        field.isTimestamp = true;
        field.serverValue = true;
        field.isRequired = true;
    }
    if (field.isTimestamp) {
        field.isBigInt = true;
    }
    if (field.isSymbol && !field.textLength) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Symbol is specified without fixed text length ${prototype.constructor.name}:${propertyKey}`);
    }
    //-- properties validation
    if (field.pk && field.fk) {
        throw Errors_1.Errors.SYSTEM_ERROR(`Field both declared as primary key and foreign key ${prototype.constructor.name}:${propertyKey}`);
    }
};
exports.Column = Column;
const FK = (config) => {
    //-- get primary key of foreign model
    let props = {};
    let relationDto;
    if (config.model) {
        relationDto = (0, metadata_1.getObjectMetadata)(config.model);
        const primaryKey = relationDto.fields.find((field) => field.pk);
        if (primaryKey) {
            props.isBigInt = primaryKey.isBigInt;
            props.textLength = primaryKey.textLength;
        }
    }
    props.fk = {
        dtoId: relationDto === null || relationDto === void 0 ? void 0 : relationDto.id,
        cascade: config.cascade,
    };
    return props;
};
exports.FK = FK;
